//
// Copyright (C) 2008 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

#include "Patricia.h"

Patricia::Patricia () 
{	
	root		= new NODE ();
	root->count = 0;
}

Patricia::~Patricia () 
{
	delete root;
}

void Patricia::insert (IP_ADDR addr, unsigned short mask)
{
	//
	// find the longest prefix
	//

	PNODE tree		= find (addr);
	PNODE newchild	= NULL;

	//
	// address already inserted?
	//

	if (tree != NULL && mask == tree->count)
		return;

	//
	// insert the remaining bits
	//

	for (int i=tree->count; i<mask; i++, tree = newchild) {
		
		unsigned short	byte	= i / 8;
		unsigned short	bit		= i % 8;

		unsigned char	byteval = addr.x [byte];
		bool			bitval	= (byteval & (1 << (8 - bit - 1))) != 0 ? true : false;
		
		newchild				= new NODE ();
		newchild->value			= bitval;
		newchild->count			= i + 1;

		if (bitval == false) {
			assert (tree->left == NULL);		
			tree->left = newchild;
		} else if (bitval == true) {
			assert (tree->right == NULL);
			tree->right = newchild;
		}

	} // for (int i=cur->count; i<mask; i++) 

}

Patricia::PNODE Patricia::find (IP_ADDR addr, PNODE tree, unsigned short cnt)
{
	//
	// first some checks 
	//

	if (tree == NULL)
		return find (addr, root, 0);

	if (tree->left == NULL && tree->right == NULL)
		return tree;

	if (tree->left != NULL)
		assert (tree->left->value == false);

	if (tree->right != NULL)
		assert (tree->right->value == true);

	//
	// which byte in the IP_ADDR are we checking?
	//

	unsigned short bytecnt = cnt / 8;
	assert (bytecnt <= 3);

	unsigned char	byte	= addr.x[bytecnt];
	unsigned short	bit		= cnt % 8;
	
	//
	// get the bit value from the IP_ADDR
	//

	bool bitvalue = (byte & (1 << (8 - bit - 1))) != 0 ? true : false;

	//
	// walk down the tree if needed
	//

	if (bitvalue == true) {
		
		if (tree->right == NULL)
			return tree;
		else
			return find (addr, tree->right, cnt + 1);

	} else {

		if (tree->left == NULL)
			return tree;
		else
			return find (addr, tree->left, cnt + 1);

	}

	//
	// should never get here
	//

	assert (false);
	return NULL;
}




